# -*- coding: utf-8 -*-

import pika
import os
import socket
import sys
import json
import time
import itertools
import traceback
import logging


class RpcError(Exception):
    pass


class RpcNamespace(object):

    def __init__(self, bus, ns=None):
        self.__bus = bus
        self.__path = ns

    def __getattr__(self, ns):
        if self.__path is None:
            path = ns
        else:
            path = '%s.%s' % (self.__path, ns)
        return RpcNamespace(self.__bus, path)

    def __call__(self, *args, **kwds):
        timeout = None
        if args:
            timeout = kwds.pop('timeout', None)
            assert not kwds, repr((args, kwds))
            params = list(args)
        elif kwds:
            params = [kwds]
        else:
            params = []
        return self.__bus.rpc_call(self.__path, *params, timeout=timeout)


class BaseBus(object):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT',
                 hostname=None, program=None):
        if hostname is None:
            hostname = socket.gethostname()
        if program is None:
            program = os.path.split(sys.argv[0])[1]
        self.url = url
        self.rpcExchange = rpcExchange
        self.eventExchange = eventExchange
        self.hostname = hostname
        self.program = program
        self.pid = os.getpid()
        self._calls = {}

    def send(self, routing_key, event, exchange=None):
        if exchange is None:
            exchange = self.eventExchange
        self.publish(
            exchange=exchange,
            routing_key=routing_key,
            body=json.dumps(event),
            mandatory=False,
        )

    def publish(self, exchange, routing_key, body,
                properties=None, mandatory=False):
        raise NotImplementedError

    def on_message(self, channel, method, props, body):
        callback, _ = self._calls.pop(props.correlation_id, (None, None))
        if callback is not None:
            callback(json.loads(body))

    def on_return(self, channel, method, props, body):
        callback, _ = self._calls.pop(props.correlation_id, (None, None))
        if callback is not None:
            body = json.loads(body)
            callback(dict(
                id=body['id'],
                error='method %s not exists' % method.routing_key,
            ))

    _nextCorrelationId = itertools.count().next

    def nextCorrelationId(self):
        return 'req-%d-%s-%d' % (
            self.pid, id(self), self._nextCorrelationId()
        )

    EXPIRATION_DELAY = 10

    DEFAULT_TIMEOUT = 300

    def time(self):
        raise NotImplementedError

    def rpc_request(self, method, *params, **kwds):
        timeout = int(kwds.get('timeout') or self.DEFAULT_TIMEOUT)
        cid = self.nextCorrelationId()
        self._calls[cid] = (kwds.get('callback'), self.time() + timeout)
        props = pika.BasicProperties(
            reply_to=self.rpcQueue,
            correlation_id=cid,
            expiration=str((timeout + self.EXPIRATION_DELAY) * 1000),
        )
        request = dict(id=cid, method=method, params=params, timeout=timeout)
        return dict(
            exchange=self.rpcExchange,
            routing_key=method,
            properties=props,
            body=json.dumps(request),
            mandatory=True,
        )

    def rpc_call(self, method, *params, **kwds):
        raise NotImplementedError

    @property
    def rpc(self):
        return RpcNamespace(self)


class Bus(BaseBus):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT',
                 hostname=None, program=None):
        super(Bus, self).__init__(
            url=url, rpcExchange=rpcExchange, eventExchange=eventExchange,
            hostname=hostname, program=program)
        self.conn = pika.BlockingConnection(pika.URLParameters(self.url))
        self.channel = self.conn.channel()
        self.queue = None

    def publish(self, exchange, routing_key, body,
                properties=None, mandatory=False):
        self.channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=body,
            properties=properties,
            mandatory=mandatory,
        )

    @property
    def rpcQueue(self):
        if self.queue is None:
            self.queue = self.channel.queue_declare(exclusive=True)
            self.channel.add_on_return_callback(self.on_return)
            self.channel.basic_consume(self.on_message, self.queue.method.queue,
                                       no_ack=True)
        return self.queue.method.queue

    def time(self):
        return time.time()

    def rpc_call(self, method, *params, **kwds):
        timeout = int(kwds.get('timeout') or self.DEFAULT_TIMEOUT)
        result = []
        kwds['callback'] = result.append
        rpc_msg = self.rpc_request(method, *params, **kwds)
        self.publish(**rpc_msg)
        while not result:
            self.conn.process_data_events(1)
        if not result:
            result = dict(
                id=rpc_msg['properties'].correlation_id,
                error="Timeouted after %s seconds" % timeout,
                result=None,
            )
        else:
            result = result[0]
        if result.get('error'):
            raise RpcError(result['error'])
        else:
            return result.get('result')

    def monitor(self, bindings, callback):
        chan = self.conn.channel()
        queue = "%s-monitor-%s-%d" % (self.program, self.hostname, self.pid)
        chan.queue_declare(
            queue=queue,
            exclusive=True,
        )
        for exchange, routing_key in bindings:
            print 'exchange=%s routing_key=%s' % (exchange, routing_key)
            chan.queue_bind(
                queue=queue,
                exchange=exchange,
                routing_key=routing_key,
                # nowait=True,
                # arguments=None
            )
        chan.basic_consume(
            consumer_callback=callback,
            queue=queue,
            no_ack=True,
            exclusive=True,
            # consumer_tag=None,
            # arguments=None
        )
        print 'start receiving messages...'
        chan.start_consuming()

    def close(self):
        if self._sendChannel is not None:
            self._sendChannel.close()
        self.conn.close()

    def startConsumer(self, bindings, queueName, callback, exchange=None):
        if exchange is None:
            exchange = self.eventExchange
        chan = self.conn.channel()
        chan.queue_declare(
            queue=queueName,
            passive=False,
            durable=False,
            exclusive=False,
            auto_delete=True,
            nowait=False,
            arguments=None
        )
        for rk in bindings:
            chan.queue_bind(
                queue=queueName,
                exchange=exchange,
                routing_key=rk,
                nowait=False,
                arguments=None
            )

        def consumer(ch, method, props, body):
            body = json.loads(body)
            callback(method.routing_key, body)
        chan.basic_consume(consumer, queue=queueName, no_ack=False)
        chan.start_consuming()

    def startRPC(self, methods, queueName, qos=1):
        chan = self.conn.channel()
        if qos:
            chan.basic_qos(prefetch_count=qos)
        chan.queue_declare(
            queue=queueName,
            passive=False,
            durable=False,
            exclusive=False,
            auto_delete=True,
            nowait=False,
            arguments=None
        )
        for name in methods:
            chan.queue_bind(
                queue=queueName,
                exchange=self.rpcExchange,
                routing_key=name,
                nowait=False,
                arguments=None
            )

        def consumer(ch, method, props, body):
            req = reqid = None
            name = method.routing_key
            cid = props.correlation_id
            # print 'REQ', method, props, body

            def reply(result=None, error=None):
                # print 'REPLY', (result, error)
                rep = dict(id=reqid)
                if result is not None:
                    rep['result'] = result
                if error is not None:
                    rep['error'] = error
                chan.basic_publish(
                    exchange="",
                    routing_key=props.reply_to,
                    properties=pika.BasicProperties(correlation_id=cid),
                    body=json.dumps(rep),
                    mandatory=False,
                    immediate=False
                )
            try:
                req = json.loads(body)
                reqid = req.get('id')
                if name not in methods:
                    reply(error='Method %s not implemented' % name)
                res = methods[name](*req['params'])
                reply(result=res)
            except Exception, e:
                reply(error='%s: %s' % (e.__class__.__name__, str(e)))
            chan.basic_ack(delivery_tag=method.delivery_tag)
        chan.basic_consume(consumer, queue=queueName, no_ack=False)
        chan.start_consuming()

ZZNBus = Bus


class TornadoBus(BaseBus):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT',
                 hostname=None, program=None):
        super(TornadoBus, self).__init__(
            url=url, rpcExchange=rpcExchange, eventExchange=eventExchange,
            hostname=hostname, program=program)
        self.conn = pika.TornadoConnection(
            pika.URLParameters(url),
            on_open_callback=self.on_connection_open
        )
        self.channel = None
        self.rpcQueue = None
        self.main = None
        self.timeoutHandler = None
        self.ioloop = self.conn.ioloop

    def on_connection_open(self, con):
        self.conn.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        self.channel = channel
        self.channel.queue_declare(self.on_queue_declared, exclusive=True)

    def on_queue_declared(self, frame):
        self.rpcQueue = frame.method.queue
        self.channel.basic_consume(self.on_message, self.rpcQueue, no_ack=True)
        self.channel.add_on_return_callback(self.on_return)
        if self.main is not None:
            self.main(self)

    def publish(self, exchange, routing_key, body,
                properties=None, mandatory=False):
        self.channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=body,
            properties=properties,
            mandatory=mandatory,
        )

    def time(self):
        return self.ioloop.time()

    def on_timeout(self):
        now = self.time()
        for cid, (callback, tout) in self._calls.items():
            if tout <= now:
                self._calls.pop(cid)
                callback(dict(
                    id=cid,
                    error='timeouted',
                ))
        self.schedule_timeout()

    def schedule_timeout(self):
        if self.timeoutHandler is not None:
            self.ioloop.remove_timeout(self.timeoutHandler)
        if not self._calls:
            return
        timeout = min(t for _, t in self._calls.values())
        self.timeoutHandler = self.ioloop.call_at(timeout, self.on_timeout)

    def rpc_call(self, method, *params, **kwds):
        from tornado.concurrent import Future
        result = Future()

        def on_rpc_result(body):
            if body.get('error'):
                result.set_exception(RpcError(body['error']))
            else:
                result.set_result(body['result'])
        kwds['callback'] = on_rpc_result
        rpc_msg = self.rpc_request(method, *params, **kwds)
        self.schedule_timeout()
        self.publish(**rpc_msg)
        return result

    def start(self, main_func):
        self.main = main_func
        self.ioloop.start()

    def stop(self):
        if self.timeoutHandler is not None:
            self.ioloop.remove_timeout(self.timeoutHandler)
        self.ioloop.stop()


AsyncBus = TornadoBus


STANDARD_ATTRS = frozenset([
    'args', 'created', 'exc_info', 'exc_text', 'filename', 'funcName',
    'levelname', 'levelno', 'lineno', 'module', 'msecs', 'msg', 'message',
    'name', 'pathname', 'process', 'processName', 'relativeCreated',
    'stack_info', 'thread', 'threadName', 'asctime',
])


def get_extra_args(logrec):
    return dict((a, v) for a, v in logrec.__dict__.items()
                if a not in STANDARD_ATTRS)


class MQHandler(logging.Handler):

    def __init__(self, bus, level=logging.NOTSET, defaults=None):
        super(MQHandler, self).__init__(level)
        if defaults is None:
            defaults = {}
        self.defaults = defaults
        self.bus = bus

    def emit(self, record):
        hostname = self.bus.hostname
        severity = record.levelname
        event = self.defaults.copy()
        msg = record.getMessage()
        if record.exc_info:
            msg = '%s\n%s' % (
                msg.strip(),
                ''.join(traceback.format_exception(*record.exc_info))
            )
        event.update({
            "timestamp": time.time(),
            "severity": severity,
            "message": msg,
            "pid": self.bus.pid,
        })
        event.setdefault('hostname', hostname)
        event.setdefault('program', self.bus.program)
        event.update(get_extra_args(record))
        self.bus.publish(
            exchange=self.bus.eventExchange,
            routing_key='log.%s.%s' % (hostname, severity),
            body=json.dumps(event),
            mandatory=False,
        )


DEFAULT_URL = "amqp://agent:{MQ_PASSWD}@mq.is.uh.in/dat"


def main():
    from argparse import ArgumentParser
    # from agent5.loggers import script_logging

    parser = ArgumentParser()
    parser.add_argument(
        '-v', '--verbose', default=0, action='count')
    parser.add_argument(
        '-U', '--url', metavar='URL', default=DEFAULT_URL,
        help='RabbitMQ URL (default %(default)s)')
    parser.add_argument(
        'action', metavar='ACTION',
        help='Action')
    parser.add_argument(
        'args', nargs='*', metavar='ARG',
        help='Action arguments')

    args = parser.parse_args()
    args.url = args.url.format(**os.environ)
    # script_logging(verbosity=args.verbose)
    # logging.getLogger('pika').setLevel(logging.WARNING)

    if args.action == 'shell':
        if args.args:
            parser.error('unwanted args for shell action')
        return action_shell(args.url)
    elif args.action == 'monitor':
        if not args.args:
            bindings = {'EVT': ['#']}
        else:
            bindings = [arg.split(':', 1) for arg in args.args]
        return action_monitor(args.url, bindings)
    else:
        parser.error('unknown action: %s' % args.action)
    return 0


def action_monitor(url, bindings):

    def dump_message(ch, method, props, body):
        print '------------------------------------------------------------'
        print '%s -> %s %s' % (
            method.exchange,
            method.routing_key,
            json.dumps(json.loads(body), indent=2, sort_keys=True)
        )
    bus = Bus(url)
    bus.monitor(bindings, dump_message)


def action_shell(url, histfile='~/.rpc-shell-history'):
    import code
    import pprint
    import rlcompleter
    import readline
    import atexit
    sys.displayhook = pprint.pprint
    bus = Bus(url)
    namespace = dict(bus=bus)
    readline.set_completer(rlcompleter.Completer(namespace).complete)
    readline.parse_and_bind("tab: complete")
    if histfile:
        histfile = os.path.expanduser(histfile)
        if os.path.isfile(histfile):
            readline.read_history_file(histfile)

        def savehist():
            readline.set_history_length(100)
            readline.write_history_file(histfile)

        atexit.register(savehist)
    con = code.InteractiveConsole(namespace)
    con.resetbuffer()
    con.interact(banner='''
bus = Bus instance (connected)
''')
